* QL WORLD DIY TOOLKIT - FAST pixel graphics routines
* Version 3.2, Copyright 1989,90,94 Simon N Goodwin.
* Modified 10/9/89 to support Thor XVI MODE 12.
* Modified 19/7/90 to suit Metacomco's Assembler.
* Modified 6/7/94 to clip points and lines in window.
* Modified 8/7/94 for line-lengths other than 128.
* Modified 15/7/94 for AH/JM ROM compatability.
* Modified 27/7/94 for Amiga Qdos and JS ROM compatability.
*
* 1994 additions feature the COHEN-SUTHERLAND LINE CLIPPING algorithm

* All coordinates are signed words, ranging from -32768 to +32767
* Use #constants not offsets for clip region limits if speed is vital

Xmin			equ			24             Window origin in MODE 4 units
Ymin			equ			26
Xmax			equ			84             Share space with Xstore
Ymax			equ			78             or Ystore, inside EXTOP DRAW
Width			equ			28					SCR/CON CHDEF A0 offset
Height		equ			30					Strictly Ysize in CHDEF
Xstore      equ         84             Offsets of old X & Y (etc.)
Ystore      equ         78             in the channel block

* Qdos System variable offsets

sv_scrst    equ         51             Byte scroll-lock indicator
sv_mcsta    equ         52             Master Chip Status shadow

* Outcode bit masks indicate which edge(s) a point is beyond

Top			equ			8       Binary masks (0 = in window)
Bottom		equ			4     	   1001    1000    1010
Right			equ			2     	   0001    0000    0010
Left			equ			1     	   0101    0100    0110

* OUTCODE macro - three register parameters: X, Y and result
* \5 is label for continuation if finished early, \4 is range

OUTCODE		macro
				moveq			#0,\3
				cmp.w			Ymax(a0),\2
				ble.s			try_Ymin\@
				moveq			#Top,\3
				bra.s			check_X\@
try_Ymin\@	cmp.w			Ymin(a0),\2
				bge.s			check_X\@
				moveq			#Bottom,\3
check_X\@	cmp.w			Xmax(a0),\1
				ble.s			try_Xmin\@
				addq.w		#Right,\3
				bra\4			\5
try_Xmin\@	cmp.w			Xmin(a0),\1
				bge\4			\5
				addq.w		#Left,\3
				endm

* CHECK_EDGE takes EIGHT parameters, and input OCODE in D2
* Assigns \7 and \8 (D2, D3) to clipped values for X and Y

CHECK_EDGE	macro
				moveq			#\6,d3
				and.w			d2,d3
				beq.s			not_this\@
				move.w		\2,\7
				sub.w			\1,\7
				move.w		\3,\8
				sub.w			\4,\8
				muls.w		\8,\7				Gives \7.L
				move.w		\5,\8
				sub.w			\4,\8
				divs.w		\8,\7				Quotient in \7.W
				add.w			\1,\7
				move.w		\3,\8
				ifnc			'\6','Left'
				bra.s			Clipping			Not needed for last case
				endc
not_this\@
				endm
*
* Initialisation for SuperBASIC extensions PLOT and DRAW
*
start       lea.l       define,a1
            move.w      $110,a2         BP.INIT vector
            jmp         (a2)
*
* Position-relative     DRAW [ # channel% , ] x% , y% 
* Absolute coordinates  DRAW [ # channel% ] TO x% , y%
*
draw        cmpa.l      a3,a5           Some parameters, please!
            beq.s       bad_param
            move.b      1(a3,a6.l),d0   Check separator flags
            moveq       #$70,d1         Trailing separator mask
            and.b       d0,d1
            cmp.b       #$50,d1         Is this a TO separator?
            beq.s       draw_to
*
* Relative DRAW uses a slightly different EXTOP routine
*
            lea.l       Reldraw,a4
            bra.s       get_params
*
* Parse  DRAW #chan TO x,y  and  DRAW TO x,y
* 
draw_to     lea.l       drawer,a4
            tst.b       d0              Does # indicate a channel?
            bmi.s       get_params
not_channel addq.l      #8,a3           Skip past the TO separator
            bra.s       get_params
*
* Handler for 2 or 3 parameters, used by PLOT and DRAW TO
*
plot        lea.l       plotter,a4
get_params  move.w      $112,a2         Vector to get integers
            jsr         (a2)            CA.GTINT
            bne.s       bad_exit
            moveq       #1,d0           First assume channel 1
            subq.w      #2,d3           At least 2 parameters?
            beq.s       get_coords      Exactly 2,  use #1
            bmi.s       bad_param       Less than 2: an error  
            subq.w      #1,d3           Only 1 parameter left?
            bne.s       bad_param       No, too many, complain
            move.w      0(a1,a6.l),d0   Get BASIC channel No.
            addq.l      #2,a1           Discard channel param.
get_coords  move.w      0(a1,a6.l),d1   Get X co-ordinate
            move.w      2(a1,a6.l),d2   Get Y co-ordinate
*
* Convert channel number in D0 to ID in A0, and call EXTOP
*
chan_sel    mulu        #40,d0          Channel table size
            add.l       $30(a6),d0      Add base offset
            cmp.l       $34(a6),d0 
            bge.s       what_chan       Past end of table?
            move.l      0(a6,d0.l),d0
            bmi.s       what_chan       Negative if closed
call_extop  move.l      d0,a0           A0 is channel ID
            move.l      a4,a2           Address of routine
            moveq       #-1,d3          Allow infinite time
            moveq       #9,d0           SD.EXTOP key
            trap        #3              Call the device driver
            rts                         Return D0 from EXTOP
*
bad_param   moveq       #-15,d0         BAD PARAMETER error
            rts
what_chan   moveq       #-6,d0          CHANNEL NOT OPEN error
bad_exit    rts                         Error code is in D0
screen_off  moveq       #-1,d0          NOT COMPLETE error
            rts
*
* RELATIVE (Spectrum-style) DRAW EXTOP; similar to DRAW TO
*
Reldraw     move.w      Xstore(a0),d5   Fetch previous X
            add.w       d5,d1
            move.w      Ystore(a0),d7   Fetch previous Y
            add.w       d7,d2
            tst.b       sv_scrst(a6)    Scroll lock?
            bne.s       screen_off      Abort if CTRL-F5 set
            btst        #3,sv_mcsta(a6) Check MODE (SV.MCSTA)
            beq.s       Mode512         MODE 4, no problems
            bclr        #0,d1           MODE 8/12, X must be EVEN
Mode512     move.l      Xmin(a0),d0
            bra.s       SavEnd
*
* Line DRAW routine; new X and Y passed from EXTOP in D1 & D2.W
*
drawer      move.w      Xstore(a0),d5   Fetch previous X
            move.w      Ystore(a0),d7   Fetch previous Y
            tst.b       sv_scrst(a6)    Test scroll lock
            bne.s       screen_off      Abort if CTRL-F5 set
            btst        #3,sv_mcsta(a6) Check MODE byte
            beq.s       Mode4           MODE 4, no problems
            bclr        #0,d1           MODE 8/12, X must be EVEN
Mode4       move.l      Xmin(a0),d0
            add.w       Xmin(a0),d1     Add window offset to X
            add.w       d0,d2           Add window offset to Y
savEnd      movem.w     d1-d2,-(a7)     Preserve end coordinates
            add.l       Width(a0),d0    Two word ADDs in one
            subq.w      #1,d0
            move.w      d0,Ymax(a0)     Store Y clipping limit
            swap        d0
            subq.w      #1,d0
            move.w      d0,Xmax(a0)     Store absolute X clip limit
*
            move.w      d1,d4           Make room in registers
            move.w      d2,d6           for Cohen-Sutherland macros
*
* The works; draw clipped line from (D4.W,D6.W) to (D5.W,D7.W)
* Uses all data registers internally; A0 -> clip limits in CHDEF

ClipLine		
				OUTCODE		d4,d6,d0,.s,GotOut0
GotOut0
				OUTCODE		d5,d7,d1,.s,ClipLoop
ClipLoop		move.w		d0,d2
				or.w			d1,d2
				beq       	draw_clip		Entirely in region
Clipper		move.w		d0,d2
				and.w			d1,d2
				bne			null_draw		Entirely outside region
				move.w		d0,d2
				bne.s			got_Outcode
				move.w		d1,d2				Outcode1 must be non zero
got_Outcode 
				CHECK_EDGE	d4,d5,Ymax(a0),d6,d7,Top,d2,d3
				CHECK_EDGE	d4,d5,Ymin(a0),d6,d7,Bottom,d2,d3
				CHECK_EDGE  d6,d7,Xmax(a0),d4,d5,Right,d3,d2
				CHECK_EDGE  d6,d7,Xmin(a0),d4,d5,Left,d3,d2
Clipping		tst.w			d0
				beq.s			done_Ocode1
				move.w		d2,d4
				move.w		d3,d6
				OUTCODE		d4,d6,d0,,ClipLoop		Set new Outcode0
				bra			ClipLoop
done_Ocode1	move.w		d2,d5
				move.w		d3,d7
				bra			GotOut0
*
draw_clip   move.l		a0,-(a7)
            moveq       #0,d0           MT.INF trap key
            trap        #1
				movea.l		(a7)+,a0
            cmp.l       #'1.10',d2      Check for old Qdos versions
            bls.s       default         JS, JM, AH, PM, EL & FB use default!
            move.w		100(a0),d2      Check SD.LINEL
            bgt.s       shuffle         Accept 1-32767, otherwise...
default     move.w      #128,d2         Use Sinclair's default
shuffle     move.w      d2,Xstore(a0)   Store length for later
            move.w      d5,d1           Shuffle registers back
            move.w      d7,d2
            move.w      d6,d5
            move.w      #$8080,d7       Set MODE 4 pixel mask
            btst        #3,sv_mcsta(a6) Check MODE (SV.MCSTA)
            beq.s       find_DX         MODE 4, no problems
            btst        #2,sv_mcsta(a6) Test Thor MODE 12 bit
            beq.s       not_Mode12
            move.w      #$C0C0,d7       Use the Thor XVI mask
not_Mode12  move.b      #$C0,d7         Mask for 8 colours
*
find_DX     suba.l      a2,a2           DX := 0
            sub.w       d4,d1
            beq.s       set_X_step
            bmi.s       neg_DX
            addq.l      #1,a2           A2 := SIGN( Delta X )
            bra.s       set_X_step
neg_DX      subq.l      #1,a2
            neg.w       d1              D1 := ABS( Delta X )
set_X_step  btst        #3,sv_mcsta(a6) Check mode (SV.MCSTA)
            beq.s       find_DY         MODE 4, D1 & A2 are OK
            adda.l      a2,a2           MODE 8/12, double X step
            asr.w       #1,d1           And halve X dot count
find_DY     suba.l      a3,a3           DY := 0
            sub.w       d5,d2
            beq.s       find_DD
            bmi.s       neg_DY
            addq.l      #1,a3           A3 := SIGN( Delta Y )
            bra.s       find_DD
*
neg_DY      subq.l      #1,a3
            neg.w       d2              D2 := ABS( Delta Y )
find_DD     cmp.w       d1,d2           D1 := MAX( D1, D2 )
            bhi.s       y_bigger
no_match    move.w      a2,a4           DDX := DX
            suba.l      a5,a5           DDY := 0
            bra.s       find_TEMP
y_bigger    exg         d1,d2           D1 := MAX( D1,D2 )
            suba.l      a4,a4           DDX := 0
            move.w      a3,a5           DDY := DY
find_TEMP   move.w      d1,d0
            asr.w       #1,d0           TEMP := STEPS DIV 2
            move.w      d2,d6           D6 := PARTS
            move.w      d1,d2
            subq.w      #1,d2           D2 is DOT for DBRA
            bmi.s       null_draw       Exit if line length=0
            exg         d1,d4           OLDX = D1; STEPS = D4
            exg         d2,d5           OLDY = D2; DOT = D5
*
* Now draw a line of D5 pixels; D1,D2 = X,Y of next point
*
dot_loop    add.w       d6,d0           TEMP := TEMP + PARTS
            cmp.w       d4,d0           Flag TEMP - STEPS
            bcs.s       use_DD
            sub.w       d4,d0           TEMP := TEMP - STEPS
            add.w       a2,d1           OLDX := OLDX + DX
            add.w       a3,d2           OLDY := OLDY + DY
            bra.s       do_dot
*
use_DD      add.w       a4,d1           OLDX := OLDX + DDX
            add.w       a5,d2           OLDY := OLDY + DDY
do_dot      movem.w     d1-d2,-(a7)     Preserve OLDX & OLDY
            move.w      Xstore(a0),d3   Line length, 2..32766
            bsr.s       plot_pixel
            movem.w     (a7)+,d1-d2     Restore OLDX & OLDY
            dbra        d5,dot_loop
null_draw   move.w      (a7)+,Xstore(a0)   Store end X on SD.XORG
            move.w      (a7)+,Ystore(a0)   Store end Y on SD.YORG
            moveq       #0,d0
            rts
*
* Clipped pixel PLOT code, uses A0, A6, D7; alters A1, D0, D1, D2, D3
*
plotter     move.w      #$8080,d7       Preset MODE 4 pixel mask
            btst        #3,sv_mcsta(a6) Check MODE (SV.MCSTA)
            beq.s       easymode        MODE 4, no problems
            bclr        #0,d1           MODE 8/12, X must be EVEN
            btst        #2,sv_mcsta(a6) Test Thor MODE 12 bit
            beq.s       not_a_thor
            move.w      #$C0C0,d7       Use the 16 colour mask
not_a_thor  move.b      #$C0,d7         Mixed mask for 8 colours
easymode    tst.b       sv_scrst(a6)    Test for scroll lock
            bne         screen_off      Abort if CTRL-F5 set
            tst.w       d2              Check Y >= 0
            bmi.s       nullplot        Off screen, no report
            cmp.w       Height(a0),d2   Check Y < CH.HEIGHT 
            bcc.s       nullplot
            tst.w       d1              Check X >= 0
            bmi.s       nullplot
            cmp.w       Width(a0),d1    Check X < CH.WIDTH
            bcc.s       nullplot
*
* To save 14 bytes at a small speed cost, BSR.S NULLPLOT here
*
				bsr.s			nullplot
*
* It's worth it to bring PLOT_PIXEL in short branch range for DRAW
*
*           add.w       Xmin(a0),d1     Add window offset to X
*           add.w       Ymin(a0),d2     Add window offset to Y
*           move.w      d1,Xstore(a0)   Store end X on SD.XORG
*           move.w      d2,Ystore(a0)   Store end Y on SD.YORG
*            
* Use SD.LINEL if it's plausible, otherwise assume 128 bytes
*
				movem.l		d1-d2/a0,-(a7)  Save our registers
            moveq       #0,d0           MT.INF trap key
            trap        #1
            cmp.l       #'1.10',d2      Check for old Qdos versions
            movem.l     (a7)+,d1-d2/a0  Restore and tidy the stack
            bls.s       old_console     JS, JM, AH, PM, EL & FB default
            move.w		100(a0),d3      Check SD.LINEL
            bgt.s       got_width       Accept 1..32767
old_console move.w      #128,d3         Use Sinclair's default
got_width   moveq       #0,d0           No error from EXTOP now
*
* D3.W is line length. Point A1 at the relevant word in video RAM
*
plot_pixel  mulu        d2,d3           Usually 128 bytes on a QL
            move.l      50(a0),a1       A1 := screen RAM start
				adda.l      d3,a1           A1 := line RAM start
            move.l      62(a0),d3       Grab the ink mask
            btst        #0,d2           Find out if Y is odd
            beq.s       find_word       It's even, mask is OK
            swap        d3              Odd, so use other mask
*      
find_word   moveq       #7,d2
            and.w       d1,d2           Save X MOD 8 in D2
            lsr.w       #3,d1           D1 := offset on line
            add.w       d1,d1           Offset := 0-126 (EVEN)
            adda.w      d1,a1           A1 -> screen RAM word
*
* Set the pixel in MODE 4, 8 or 12; allow for OVER 0, 1 or -1
*
set_pixel   move.w      d7,d1           Select pixel mask
            ror.w       d2,d1           Align pixel mask
            and.w       d1,d3           D3 := 1 pixel of ink
            btst        #3,66(a0)       Check for OVER -1
            bne.s       overplot        OVER 0 matches OVER 1
            not.w       d1              Mask out the rest
            and.w       (a1),d1         Fetch the background
            or.w        d3,d1           Insert the pixel
            move.w      d1,(a1)         Store the combination
            rts                         PLOT returns ERR.OK
overplot    eor.w       d3,(a1)         Exclusive OR the pixel
            rts
*
* Exit, updating the absolute co-ordinates corrected for MODE
*
nullplot    add.w       Xmin(a0),d1     Add window offset to X
            add.w       Ymin(a0),d2     Add window offset to Y
            move.w      d1,Xstore(a0)   Store end X on SD.XORG
            move.w      d2,Ystore(a0)   Store end Y on SD.YORG
            rts
*
define      dc.w        2               Number of procs
            dc.w        plot-*
            dc.b        5,'XPLOT'
            dc.w        draw-*
            dc.b        5,'XDRAW'
            dc.w        0,0,0           End of procs, no fns

				end

