* QL World DIY Toolkit, HPDUMP for Qdos screens
* Version 2.18m, Copyright Simon N Goodwin March/April 1993
* Updated September 1993 to suit Metacomco Assembler v 1.7
*
* HPDUMP_TASK runs immediately after CALL or EXEC:
*
* CALL parameter controls black<->white; 0 for true <>0 for inverse
* i.e. CALL x,1 gives inverse; default CALL x gives a true dump.
*
* If EX(EC){UTE}ed, a one-chararacter parameter string "!" signals
* an inverse image; longer strings are taken as a device name, with
* an optional "!" on the end, stripped before the names is OPENed.
* Optional channel parameter is the output ID, left open after use.
*
* The HPDUMP_CODE version adds HPDUMP [!] as a resident procedure.
* The optional following ! asks HPDUMP for an inverse image.
* A string parameter or channel number prefixed by # may preceed
* the optional !, overriding the default output device name.
*
* If two screens are supported HPDUMP uses the one currently shown.
*
* Register utilisation:
*
* D7 is 0 for transient (CALL/RESPROC) usage with an ALCHP buffer,
* 1 for EX opening a temporary channel, -1 for EX #channel%
* D5 is -1 for INVERSE and 0 for TRUE shaded black on white paper.
* A0 is the output channel ID, sometimes moved to D6 temporarily.
* While generating a dump:
* A5 points to a 128 byte buffer for the current output line.
* A2 points to a 256 word table with zeros inserted in all even bits.
* A4 points to the QL screen.
*
extension   equ      0                 1 = resident, 0 = transient
*
            ifeq     extension
*
            data     660               Task buffer and stack space
*
task_start  bra.s    multidump         6 bytes for initialisation
            dc.b     '2.18'            Version number fills four
            dc.w     $4AFB             Task header
            dc.w     6                 Name length
            dc.b     'HPDUMP'          Task name
*
multidump   lea.l    task_start,a0     Where am I ?
            cmpa.l   a0,a6             If A6 is here, I'm a task
            bne.s    calldump
            moveq    #1,d7             Flag task invocation
            move.w   (a7)+,d0          Are any channels supplied?
            beq.s    string_par
            moveq    #-15,d3           Preset error code ERR.BP
            subq.w   #1,d0
            bne.s    remove_task       Only one channel, please!
            moveq    #-1,d7            Flag: channel passed to task
            movea.l  (a7)+,a0          Fetch the channel ID
*
string_par  moveq    #0,d5             Presume not inverse
            move.w   (a7),d0           Any parameter string?
            ble.s    default_ch
strung      cmpi.b   #'!',1(a7,d0.w)   Check the end
            bne.s    not_inverse
            moveq    #-1,d5            Set inverse mask
            subq.w   #1,(a7)           Correct length
            beq.s    default_ch        No name left
not_inverse tst.l    d7                Do we have a channel yet?
            bmi.s    channel_set       If so, no need to OPEN one
            movea.l  a7,a0             Point A0 at parameter$
*
* N.B. This only works for tasks; A7 must be static (i.e. not BASIC)
*
            bra.s    chosen_ch         
*
default_ch  tst.l    d7                Is ID known yet?
            bmi.s    got_channel
            lea.l    printname,a0      Point at default name
chosen_ch   moveq    #-1,d1            Own channel
            moveq    #2,d3             New exclusive device
            moveq    #1,d0             IO.OPEN
            trap     #2
            move.l   d0,d3             Copy and test error code 
            bne.s    remove_task       How disappointing
            moveq    #1,d0
            add.w    (a7)+,d0          Round up length
            bclr     #0,d0             Keep result even
            adda.l   d0,a7             Tidy name from stack
*
got_channel lea.l    0(a6,a4),a5       Point A5 at task data area
            move.l   a0,d6             Save channel ID
            bsr.s    set_buffer
            move.l   d0,d3             Check error code
*
* D3 is the error code reply for the calling task
*
remove_task moveq    #-1,d1
            moveq    #5,d0             MT.FRJOB
            trap     #1                There's no return from here!
*
calldump    moveq    #-1,d5            Inverse mask (true brightness)
            tst.l    d1                Is parameter non-zero
            bne.s    open_out
            moveq    #0,d5             Non-inverting mask
*
            endc     call/task
*
* The BP.INIT code is only needed if this is an extension file
*
            ifne     extension
*
start       lea.l    define,a1
            movea.w  $110.w,a0         BP.INIT vector
            jmp      (a0)
*
* If HPDUMP has a "!" separator the INVERSE dump is selected.
* # signals a SuperBasic channel number, otherwise a device name.
*
hpdump      moveq    #0,d5             Non-inverting maybe?
            cmpa.l   a3,a5             Are there any parameters?
            beq.s    open_out
            moveq    #%111000,d0       Mask out separator
            and.b    1(a3,a6.l),d0
            cmp.b    #%100000,d0       Is it an exclamation mark?
            bne.s    no_shriek
            moveq    #-1,d5            Inverting mask
no_shriek   tst.b    1(a3,a6.l)        Is there a # prefix?
            bmi.s    hash_prefix
*
not_channel movea.w  $116.w,a0         CA.GTSTR vector
            jsr      (a0)
            bne.s    bad_return
            subq.w   #1,d3             One parameter?
            bne.s    bad_param
            movea.l  a1,a0             Point A0 at the string
            trap     #4                String is A6 relative
            bra.s    open_param        Try to open it
*
hash_prefix movea.w  $112.w,a0         CA.GTINT vector
            jsr      (a0)
            bne.s    bad_return
            subq.w   #1,d3             One parameter?
            bne.s    bad_param
            move.w   0(a1,a6.l),d0     Fetch parameter word
            bmi.s    bad_param
            mulu     #40,d0            Scale for BV.CH entries
            add.l    48(a6),d0         Add BV.CHBAS offset
            cmp.l    52(a6),d0         Check against BV.CHP
            bge.s    bad_channel
            move.l   0(a6,d0.l),d0     Fetch channel ID from table
            bpl.s    channel_ok
*
bad_channel moveq    #-6,d0            Channel not open
            rts
bad_param   moveq    #-15,d0
bad_return  rts
*
channel_ok  movea.l  d0,a0             Move ID to usual register
            bra.s    channel_set
            endc     extension
*
* Open the output channel 
*
open_out    lea.l    printname,a0
open_param  moveq    #0,d7             Indicate transient type
            moveq    #-1,d1            Own channel
            moveq    #0,d3             Old exclusive device
            moveq    #1,d0             IO.OPEN
            trap     #2
            tst.l    d0                Test error code 
            bne      bad_exit
*
channel_set move.l   a0,d6             Save channel ID
allocate    move.l	#128+256*2,d1		640 bytes of buffer
            moveq    #-1,d2            Owned by this job
            moveq    #24,d0
            trap     #1						MT.ALCHP
            move.l   d0,d4             Copy and test error code
            bne      tidy_up           Out of memory!
            movea.l  a0,a5             A5 -> Line buffer base
set_buffer  lea.l    128(a5),a2        A2 -> Lookup table
*
* Locate the screen
*
            movea.l  #131072+32768,a4  Assume screen 1 for now
            moveq    #0,d0
            trap     #1                MT.INF
            move.b   52(a0),d0         Test SV.MCSTAT
            bmi.s    screen1           Top bit means screen 1
            lea.l    -32768(a4),a4     Wind back to screen 0
screen1     movea.l  d6,a0             Restore channel ID     
*
* Send initialisation sequence
*
            moveq    #-1,d3            Infinite timeout
            lea.l    setup150,a1       Point at the string
            move.w   (a1)+,d2          Byte count
            bsr      send_check        
*
* Fast table maker v2.1 by Simon N Goodwin & Philip R Spink
* 38 bytes of code +16 data bytes; 60 per cent shorter and
* three times as fast as v1.
*
            movea.l  a2,a3            Copy table pointer
				lea.l    nibbles,a1		  Point at nibble data
*
* A5 -> line buffer, A3 -> pixel table, A1 -> Nibble table
*
				moveq    #15,d0           Outer loop counter
high_loop   moveq    #15,d1           Inner loop counter
low_loop    move.w   d0,d2            Form target index in D2
            lsl.w    #4,d2
            add.w    d1,d2            D2 := D0 * 16 + D1
            add.w    d2,d2            D2 is a word index
            move.b   0(a1,d0.w),d4
            lsl.w    #8,d4            D0 points at high byte
            add.b    0(a1,d1.w),d4    D1 points at low byte
            move.w   d4,0(a3,d2.w)    Store combined word
            dbra     d1,low_loop
            dbra     d0,high_loop
*
* Print out the screen addressed by A4
*
            move.w   #255,d6           Line count
do_screen   moveq    #63,d0            Do 64 words
            movea.l  a5,a3             Build line at A3
do_line     moveq    #0,d1
            move.b   (a4)+,d1          Green byte
            add.w    d1,d1             Make word index
            move.w   0(a2,d1.w),d2
            add.w    d2,d2             Move odd to even bits
            moveq    #0,d1
            move.b   (a4)+,d1          Red byte
            add.w    d1,d1             Make word index
            or.w     0(a2,d1.w),d2     Mix red & green data
            eor.w    d5,d2             Apply INVERSE/TRUE mask
            move.w   d2,(a3)+          Buffer one word
            dbra     d0,do_line
            bsr.s    send_line
*
            lea.l    -128(a4),a4       Go back again
            moveq    #63,d0            Do 64 words
            movea.l  a5,a3             Build line at A3
do_line2    moveq    #0,d1
            move.b   (a4),d1           Green byte
            add.w    d1,d1             Make word index
            move.w   0(a2,d1.w),d2     Get the odd bits
            moveq    #0,d1
            move.b   (a4)+,d1          Green again
            and.b    (a4)+,d1          Green AND red?
            add.w    d1,d1             Make word index
            move.w   0(a2,d1.w),d1     Get the other bits
            add.w    d1,d1             Make them even
            or.w     d1,d2             Combine odd and even
            eor.w    d5,d2
            move.w   d2,(a3)+
            dbra     d0,do_line2
            bsr.s    send_line
            btst     #0,d6             Skip one line in six
            bne.s    skip_line
*
* Send the final line of dots for this QL pixel line
*
            lea.l    -128(a4),a4       Go back again
            moveq    #63,d0            Do 64 words
            movea.l  a5,a3             Build line at A3
do_line3    moveq    #0,d1
            move.b   (a4)+,d1          Green byte
            add.w    d1,d1             Make word index
            move.w   0(a2,d1.w),d2     Get the odd bits
            moveq    #0,d1
            move.b   (a4)+,d1          Red byte
            add.w    d1,d1             Make word index
            move.w   0(a2,d1.w),d1     Get the other bits
            add.w    d1,d1             Make them even
            or.w     d1,d2             Combine odd and even
            eor.w    d5,d2
            move.w   d2,(a3)+
            dbra     d0,do_line3
            bsr.s    send_line
*
skip_line   dbra     d6,do_screen
            moveq    #0,d4             ERR.OK, it worked
*
close_out   tst.l    d7                Check EXEC/CALL flag
            bmi.s    d4_error          No CLOSE if task parameter
            bne.s    tidy_up           No RECHP for tasks 
            exg      a5,a0
            moveq    #25,d0            MT.RECHP
            trap     #1
            exg      a5,a0
tidy_up     moveq    #2,d0            	IO.CLOSE
            trap     #2
d4_error    move.l   d4,d0             Return error code
bad_exit    rts
*          
* Write the line of pixels at A5 to channel A0
*
send_line   lea.l    newline,a1
            move.w   (a1)+,d2          Fetch length
            bsr.s    send_string       Initialise line
            bne.s    fall_out
            movea.l  a5,a1             Point at the bytes
            move.w   #128,d2           Byte count
send_check  bsr.s    send_string
            beq.s    exit_ok
fall_out    addq.l   #4,a7             Discard return address
            bra.s    close_out
*
send_string moveq    #7,d0             IO.SSTRG
            trap     #3
            move.l   d0,d4             Set Z flag if it worked
exit_ok     rts
*
* Table of values 0 to 15 in even bits, with odd bits zero
*
nibbles     dc.b     0,1,4,5,16,17,20,21,64,65,68,69,80,81,84,85
*
* HP PCL escape sequences, at the end for easy patching
*
printname   dc.w     4
            dc.b     'SER1'
newline     dc.w     7
            dc.b     27,'*b128W'       128 bytes per line
setup150    dc.w		15
            dc.b		27,'*t150R'       150 DPI density
            dc.b     27,'*r1024S'      1024 pixels per line
*
            ifne     extension
*
define      dc.w     1                 One procedure
            dc.w     hpdump-*
            dc.b     6,'HPDUMP'
            dc.w     0,0,0             That's all
*
            endc     extension
*
            end

