* QL WORLD DIY TOOLKIT - MULTIBASIC dynamic swapping routines
* Version 4.2        Copyright 1988-91        Simon N Goodwin 
*
* Updates from v 3.2 and v 3.3 are marked ** 3.3 ** etc.
*
* 4.2 preserves the ALCHP chain, thanks to Laurence Reeves.
*
* 4.1 can save and restore the main SCREEN (32K) and MODE in
*     a MULTIBASIC file. Control this with the new command:
*
*     SCR_SAVE <flag>
*
*     FLAG  =  0   prohibits screen-saving
*     FLAG >=  1   means UNLOAD & RESAVE save the screen
*     FLAG <= -1   means save screens but do not restore
*
* SCR_SAVE -1 is a useful temporary instruction when you want
* to RELOAD a MultiBASIC saved with its screen, without losing
* information on the current display.
*
* Notes:
*
* (0) Adjust the equate SCRBASE, below, if video RAM does
*     not start at $20000 on your system.
*
* (1) The 'code' area in MultiBASIC 4.0 tasks is expanded
*     to include an extra word, at the end, containing the
*     value of the SCR_SAVE flag at the time of UNLOADing. 
*
* (2) MULTIBASIC cannot save more than 512K correctly in
*     one task. This limit is unlikely to cause problems!
*
*
looper  equ     24830           Opcode of BRA.S *-2
scrbase equ     131072          Start of 32K screen RAM
*
start   lea.l   define,a1
        move.w  $110,a2         BP.INIT vector
        jsr     (a2)            Add new commands
        lea.l   hello,a1        ** 4.0 ** Sign on
        suba.l  a0,a0           ** 4.0 ** Channel (0,0)
        move.w  $D0,a2          ** 4.0 ** UT.MTEXT
        jmp     (a2)            ** 4.0 **
*  
hello   dc.w    64              ** 4.0 **
        dc.b    'QDOS MULTIBASIC extensions v4.2',10
        dc.b    'Copyright 1988-91 Simon Goodwin',10
*
scrsave move.w  $112,a2         ** 4.0 ** CA.GTINT
        jsr     (a2)            ** 4.0 **
        bne.s   dropout         ** 4.0 **
        subq.w  #1,d3           ** 4.0 **
        bne.s   bad_par         ** 4.0 **
        lea.l   scrflag,a0      ** 4.0 **
        move.w  0(a6,a1.l),(a0) ** 4.0 **
        moveq   #0,d0           ** 4.0 **
        rts                     ** 4.0 **
*         
*       SHARED KEYWORD CODE 
*
unload  moveq   #-1,d7          Flag for UNLOAD
        bra.s   getname
reload  moveq   #1,d7           Flag for RELOAD
        bra.s   getname
resave  moveq   #-2,d7          Flag for RESAVE
        bra.s   getname
remove  moveq   #0,d7           Flag for REMOVE
*
getname lea     8(a3),a4        Check for one parameter
        cmpa.l  a4,a5
        beq.s   one_par
bad_par moveq   #-15,d0         Report BAD PARAMETER
dropout rts
one_par tst.b   0(a3,a6.l)      Does it have a value?
        beq.s   unset           No, use the name instead
        move.w  $116,a0         CA.GTSTR vector
        jsr     (a0)            Get the value as a string
        bne.s   dropout         Give up if that went wrong
        move.l  $58(a6),a1      Get string offset (BV.RIP)
        tst.b   0(a1,a6.l)      Check length is 0-255
        bne.s   bad_par         No good, length >= 256
        lea.l   1(a1),a5        0(A5,A6.L) -> length byte
        bra.s   tst_len
*
unset   move.l  24(a6),a0       Find BASIC's Name Table
        move.w  2(a3,a6.l),d0   Index of actual parameter
        lsl.w   #3,d0           N.T. entries take 8 bytes
        adda.w  d0,a0           (A0,A6.L) -> N.T. Entry
        move.w  2(a0,a6.l),d0   D0 is name offset in list
        move.l  32(a6),a5       Find BASIC's Name List
        adda.w  d0,a5           0(A5,A6.L) -> length byte
tst_len tst.b   0(a5,a6.l)      ** 3.3 ** Allow 1-127 chars
        ble.s   bad_par
*
* Look up the required name amongst all the tasks
*
look_up moveq   #0,d1           Start with task 0,0
        bra.s   scan            Scan all the tasks
look_on tst.l   d1              At end of task tree?
        beq.s   do_cmd          If so, do the command
scan    moveq   #0,d2           Scan from top of tree
        move.l  d1,d4           Save current task ID
        moveq   #2,d0           MT.JINF trap key
        trap    #1              Find the next task
        tst.l   d0              Does it exist?
        bne.s   do_cmd          If not, do the command
*
* A0 -> task, D4 is task ID, D1 is ID of 'next' task
*
        move.l  a0,d6           Save base address of task
        addq.l  #2,a0           Skip BRA.S at the start
        move.l  (a0)+,d5        Get length
        cmpi.w  #$4AFB,(a0)+    Does task have a header?
        bne.s   look_on         No, don't look for a name
        tst.b   (a0)+           Name length should be <256
        bne.s   look_on         Otherwise, keep looking
        move.b  0(a5,a6.l),d0   Get parameter name length
        cmp.b   (a0)+,d0        Do name lengths match? 
        bne.s   look_on         No, try another task
        move.l  a5,a4           Copy name pointer for test
compare move.b  1(a4,a6.l),d2   Get parameter character
        moveq   #-33,d3         Mask to ignore letter case
        and.b   d3,d2           Convert parameter to caps         
        and.b   (a0)+,d3        Convert header to capitals
        cmp.b   d2,d3           Compare characters
        bne.s   look_on         Look on unless they match
        addq.l  #1,a4           Advance through parameter
        subq.b  #1,d0           Count one less to check
        bne.s   compare         Check all the characters
        moveq   #-1,d1          Flag that name was found
*
* If D1 is negative, task was found: code at D6, length D5
*
do_cmd  tst.l   d7              What operation is needed?
        bmi.s   unloda          Negative, UNLOAD / RESAVE
        bne     reloda          Positive, RELOAD
*
* REMOVE routine - if we found a task, we can remove it
*
remova  tst.l   d1              Was the name found?
        bmi.s   kill_it
bad_job moveq   #-2,d0          Report INVALID JOB
        rts
exists  moveq   #-8,d0          Report ALREADY EXISTS
        rts
unknown moveq   #-7,d0          Report NOT FOUND
        rts
*
kill_it move.l  d4,d1           Set up task ID
        moveq   #5,d0           MT.FRJOB trap key
        trap    #1              Remove the task
        rts                     Report back to caller
*
* UNLOAD routine - copies BASIC to an image in task RAM
*
unloda  tst.l   d1              Did we find the name?
        bpl.s   make_it         No, so create it
        addq.w  #1,d7           Is this UNLOAD?
        beq.s   exists          If so, it already exists
        bsr.s   kill_it         Otherwise it's RESAVE...
make_it moveq   #0,d0           MT.INF trap key
        trap    #1              Find system variables
        trap    #0              Stop anything moving!
        ori.w   #$700,sr        Disable interrupts
        move.l  a0,a3           A3 -> system variables
        move.l  20(a0),d3       D3 -> End of BASIC+1
        sub.l   16(a0),d3       D3 = Length of BASIC area
*
* D3 is length of 'data' area. Get 'code' length into D2.
*
        moveq   #20,d2          ** 4.0 ** D2 = Minimum 'code'
        move.b  0(a5,a6.l),d5   Find length of task name
        addq.b  #1,d5           ... and ensure it is even
        bclr    #0,d5           Round length to even
        add.b   d5,d2           Add name to 'code' length
*
** 4.0 ** Check SCR_SAVE flag
*
        move.w  scrflag,d0      ** 4.0 ** Are we saving screens?
        beq.s   alloc8          ** 4.0 ** If not, continue
        add.l   #32770,d3       ** 4.0 ** Add space for screen & mode
*
** 4.0 **
*
* Allocate transient program memory for the image
*
alloc8  moveq   #0,d1           Owner: SuperBASIC
        move.l  d1,a1           Put task in TRNSP
        moveq   #1,d0           MT.CJOB trap key
        trap    #1              Create a transient task
        tst.l   d0              Check for errors
        bne.s   unlock          Give up if CJOB failed
*
** 4.0 ** Check SCR_SAVE flag
*
        move.w  scrflag,d0      ** 4.0 ** Are we saving screens?
        beq.s   no_scr          ** 4.0 ** If not, continue
        sub.l   #32770,d3       ** 4.0 ** Make D3 size of BASIC only
no_scr
*
** 4.0 **
*
* Set up a dummy task header
*
        move.w  #looper,(a0)+   START is BRA.S START !
        move.l  d3,(a0)+        Store length AFTER header
        move.w  #$4AFB,(a0)+
        adda.l  a6,a5           A6 cannot change now
        move.b  (a5)+,d0        Get length of task name
        ext.w   d0
        move.w  d0,(a0)+        Store length (always >0)
setname move.b  (a5)+,(a0)+     Copy name a byte at a time
        subq.b  #1,d5           Use even count in D5 to
        bne.s   setname         ... ensure A0 ends up even
        move.l  a6,(a0)+        Save base address of BASIC
        move.l  usp,a5          Save User Stack Pointer
        move.l  a5,(a0)+
*
        move.l  a0,a2           ** 4.0 ** Save flag copy pointer
        move.w  scrflag,(a0)+   ** 4.0 ** Save current SCR_SAVE flag
        move.l  224(a6),d1      ** 4.2 ** Save ALCHP sentinel
*
        move.l  16(a3),a3       A3 -> NEW start of BASIC
copy_d3 lsr.l   #3,d3           Count pairs of long words
        subq.l  #1,d3           Adjust for DBRA loop
copier  move.l  (a3)+,(a0)+     It's fast and safe to move
        move.l  (a3)+,(a0)+     ... eight bytes at a time
        dbra    d3,copier       Move 270-540 QL K a second
        move.l  d1,224(a6)      ** 4.2 ** Restore ALCHP sentinel
*
*
** 4.0 ** Save screen, if necessary
*
        tst.w   (a2)            ** 4.0 ** Test copy of flag
        beq.s   lockout         ** 4.0 ** Job done, return
        moveq   #-1,d1          ** 4.0 ** Read current MODE
        moveq   #-1,d2          ** 4.0 ** Ignore display type
        moveq   #16,d0          ** 4.0 ** MT.DMODE
        trap    #1              ** 4.0 **
        move.w  d1,(a0)+        ** 4.0 ** Save mode before screen
        lea.l   scrbase,a3      ** 4.0 **
copy32  move.w  #4095,d5        ** 4.0 ** DBRA count for 32K
copy2   move.l  (a3)+,(a0)+     ** 4.0 **
        move.l  (a3)+,(a0)+     ** 4.0 **
        dbra    d5,copy2        ** 4.0 **
*
** 4.0 ** End of screen saver
*
lockout moveq   #0,d0           Report no error
unlock  andi.w  #$D8FF,sr       Re-start QDOS
        rts
* 
* RELOAD <name>  - fetch task from (D6), length in D5
*
reloda  tst.l   d1              Did we find the name?
        bpl     unknown         ** 4.0 ** formerly BPL.S
load_it move.b  0(a5,a6.l),d7   Find name length for later
        addq.b  #1,d7
        bclr    #0,d7           Round length to even
        moveq   #0,d0           MT.INF Trap key
        trap    #1              Find system variables
        move.l  a0,a4           Save system var. pointer
        move.l  20(a0),d1       D1 -> End of BASIC+1
        sub.l   16(a0),d1       D1 := Length of BASIC area
        sub.l   d5,d1           D1 is delta size
        beq.s   size_ok         No change required
        bpl.s   shrink          Discard excess space
        neg.l   d1              Expand by D5-D1 bytes
        moveq   #22,d0          MT.ALBAS trap key
        trap    #1              Expand BASIC work space
        tst.l   d0              Check for OUT OF MEMORY
        beq.s   size_ok         
        rts
shrink  moveq   #23,d0          MT.REBAS trap key
        trap    #1              Shrink BASIC work space
*
* Now work out the address of the code in the task space
*
size_ok moveq   #10,d2          Skip fixed part of header
        add.b   d7,d2           Skip name (max length 127)
        add.l   d2,d6           Point D6 at saved A6 & A7
        move.l  d6,a3           Set source pointer
        trap    #0              Stop BASIC space moving
        ori.w   #$700,sr        Disable interrupts
*
        move.l  a6,a0           Offset user A7 by delta A6
        suba.l  (a3)+,a0        Delta address of BASIC
        adda.l  (a3)+,a0
        move.l  a0,usp          USP is delta A6 + old USP
*
        move.l  a3,a2           ** 4.0 ** Save flag pointer
        addq.l  #2,a3           ** 4.0 ** Skip over flag
*
        move.l  16(a4),a0       Get destination, SV.BASIC
        lsr.l   #3,d5           Count long word pairs
        subq.l  #1,d5           Adjust for DBRA
copy    move.l  (a3)+,(a0)+     Move task data to BASIC
        move.l  (a3)+,(a0)+
        dbra    d5,copy
*
** 4.0 ** Restore screen & MODE, if necessary
*
        tst.w   (a2)            ** 4.0 ** Test saved flag
        beq.s   lockout         ** 4.0 ** Job done, return
        move.w  scrflag,d1      ** 4.0 ** Check new orders...
        bmi.s   lockout         ** 4.0 ** Temporary over-ride
        moveq   #-1,d1          ** 4.0 ** Read current MODE
        moveq   #-1,d2          ** 4.0 ** Ignore display type
        moveq   #16,d0          ** 4.0 ** MT.DMODE
        trap    #1              ** 4.0 **
        cmp.w   (a3)+,d1        ** 4.0 ** New MODE setting?
        beq.s   mode_ok         ** 4.0 ** If not, just restore
        move.w  -2(a3),d1       ** 4.0 ** Re-read old mode
        moveq   #16,d0          ** 4.0 ** MT.DMODE
        trap    #1              ** 4.0 ** Set to old mode
mode_ok lea.l   scrbase,a0      ** 4.0 ** Source address
        bra     copy32          ** 4.0 **
*
* SCR_SAVE flag; Move this if putting MultiBASIC in ROM!
*
scrflag dc.w    1               ** 4.0 ** Default YES to SCR_SAVE
*
define  dc.w    5               ** 4.0 ** No. of procedures (was 4)
        dc.w    scrsave-*
        dc.b    8,'SCR_SAVE'
*
** 4.0 ** End of extra data & code for SCR_SAVE command
*
        dc.w    unload-*
        dc.b    6,'UNLOAD'
        dc.w    reload-*
        dc.b    6,'RELOAD'
        dc.w    resave-*
        dc.b    6,'RESAVE'
        dc.w    remove-*
        dc.b    6,'REMOVE'
        dc.w    0,0,0           No functions
        end
