* QL WORLD DIY TOOLKIT - QDOS character CLIP functions
* Version 1.1a, Copyright 1990 Simon N Goodwin.
*
* CLIPBOARD functions read characters direct from any QDOS SCR or CON
* window. Recognises all colours and all distinct stipples, standard
* & extra CSIZEs (e.g. 0,0  3,1 or 7 pixel Psion '64 column' mode).
* Scans 9x6 pixel grid for each character.
* Works with all ASCII character codes 0-255, except 10, which is
* not recognisable on QDOS (or Unix, etc) as it's always invisible!
* 
start      lea.l    define,a1
           move.w   $110,a2         BP.INIT vector
           jmp      (a2)
*
* CLIP% & CLIP$ code - process channel parameter
*
clip_int   moveq    #-1,d7          Flag integer return
           bra.s    clip_param
clip_str   moveq    #0,d7           Flag string result
clip_param move.w   $118,a2         CA.GTLIN Vector, ensures
           jsr      (a2)            4 bytes of stack space
           bne.s    bad_exit
           subq.w   #1,d3           One parameter expected
           bne.s    bad_param       
           move.l   0(a1,a6.l),d0   Get BASIC channel number
           addq.l   #2,a1           Leave room for one INT
           move.l   a1,$58(a6)      Set maths stack BV.RIP
*
* Check and convert channel number in D0 to ID in A0
*
chan_sel   mulu     #40,d0          Channel table size
           add.l    $30(a6),d0      Add base offset
           cmp.l    $34(a6),d0 
           bge.s    what_chan       Past end of table?
           move.l   0(a6,d0.l),d0
           bmi.s    what_chan       Negative if closed
chan_open  move.l   d0,a0           A0 is channel ID
*
* D0=9 (SD.EXTOP), A0=Channel ID, A2 = Address of CHAXTOP
* D3=0 gives ERR.NC if window in use; D3=-1 waits indefinitely
* A1 (in/out) is address of 9 byte buffer for scanned pattern
*
           lea.l    chaxtop,a2      Address of CH routine                 
           moveq    #0,d3           Don't wait around
           lea.l    -10(a7),a7      Protect some stack space
           movea.l  a7,a1           A1 -> Pattern buffer start
           moveq    #9,d0           SD.EXTOP key
           trap     #3              Call the device driver
           lea.l    10(a7),a7       Discard pattern (this time)
           move.l   $58(a6),a1      Retrieve maths stack BV.RIP
*
* If D7.L is positive return a string, otherwise an integer
* D1 should be ASCII code or -7 if NOT FOUND; D0 may be ERR.OK
* (OK), ERR.BP (not SCR/CON), ERR.OR (no room) or ERR.NC (busy)
*
good_try   tst.l    d0              Channel error?
           bpl.s    return_val
           move.l   d0,d1           Return error code in D1      
           moveq    #0,d0           Disregard EXTOP error code
return_val tst.l    d7              Integer or string result?
           bmi.s    return_int    
return_str tst.w    d1              Did the lookup succeed?
           bmi.s    return_nul      Return null "" if not
           move.b   d1,0(a1,a6.l)   Stack character code
           subq.l   #2,a1           Room for string length
           move.l   a1,$58(a6)      Update BV.RIP
           move.w   #1,0(a1,a6.l)   Stack length
           bra.s    got_str
*
return_nul clr.w    0(a1,a6.l)      Stack a null string
got_str    moveq    #1,d4           Indicate string result
           rts
*
* Result 0-255 = ASCII code of character in window; or -1 =
* Channel Busy, -4=Out of Range, -7=Not found, -15=Not SCR/CON
*
return_int move.w   d1,0(a1,a6.l)   Put result in space
           moveq    #3,d4           Indicate type is INT
           rts                      Return EXTOP error code
*
* Error return points
*
what_chan  moveq    #-6,d0          CHANNEL NOT OPEN error
bad_exit   rts                      Error code is in D0
bad_param  moveq    #-15,d0         BAD PARAMETER error
           rts
range_err  moveq    #-4,d0          Out of Range report code
           rts                      (i.e. no room for cursor)
*
* CHAracter eXTended OPeration routine: reads characters
* from display memory at the location and scale indicated by
* the current cursor parameters of the channel concerned.
* Puts an ASCII result (0-255) in D1, -7 if pattern unmatched.
* D0 error codes -1 & -4 mean the channel is busy or no room.
*
* First exclude MEM and check there's room for a character
*
chaxtop    moveq    #40,d0          Size of a MEM block
           cmp.l    (a0),d0         Exclude EXTOP to MEM as
           beq.s    bad_param       it does not suit CHAXTOP
           move.l   34(a0),d0       D0.L := XPOS & YPOS
           bmi.s    range_err       Reject if XPOS < 0
           tst.w    d0              Test Y (in lower word)
           bmi.s    range_err       Reject if YPOS < 0
           add.l    38(a0),d0       D0 := D0 + XINC & YINC
           cmp.w    30(a0),d0       Is the window too short?
           bhi.s    range_err
           swap     d0              Move X to the lower word
           cmp.w    28(a0),d0       Is the window too narrow?
           bhi.s    range_err
*
* OK, find absolute pixel co-ordinates, screen line & buffer
*
           move.w   34(a0),d1       D1 := X
           add.w    24(a0),d1       Add window offset to X
           move.w   36(a0),d2       D2 := Y
           add.w    26(a0),d2       Add window offset to Y
           movea.l  50(a0),a2       Get screen RAM base address       
           lsl.w    #7,d2           1 line uses 2^7 = 128 bytes
           adda.w   d2,a2           Get address of start of line
*
* Set up a pixel mask to suit any QL or Thor display mode
*
           move.w   #$8080,d7       Presume MODE 4 pixel mask
           btst     #3,52(a6)       Check MODE (SV.MCSTA)
           beq.s    mask_ok         MODE 4, no worries
           bclr     #0,d1           MODE 8 or 12, X must be EVEN
           btst     #2,52(a6)       Test Thor MODE 12 bit
           beq.s    not_a_thor      This must be the QL's MODE 8
           move.w   #$C0C0,d7       Use the Thor 16 colour mask
not_a_thor move.b   #$C0,d7         Adjust MODE 4 mask for 8 colours
*
* Work out the X and Y steps
* 
mask_ok    movea.w  #128,a5         Depth of one QDOS video line
           btst     #4,66(a0)       Double height attribute?
           beq.s    no_taller
           adda.l   a5,a5           Double Y step to suit
no_taller  moveq    #1,d5
           btst     #6,66(a0)       Double width attribute?
           beq.s    no_wider
           moveq    #2,d5           Boost X step
*
* At last it's time to look up a character from screen RAM
*
* Character scanner - register utilisation:
*
* D0 - Scratch
* D1 - Current pixel X co-ordinate
* D2 - line count, 8-0 top to bottom
* D3 - column count, 5-0 left to right
* D4 - Pixel bit offset in video word, 0-7 left to right
* D5 - X step between pixels: 1=CSIZE 0/1,?; 2=CSIZE 2-3,?
* D6 - Accumulated character form for this line
* D7 - Pixel mask for this MODE (WORD)
*
* A0 - Pointer to channel definition block
* A1 - Pointer to byte buffer of 9 lines in QL fount format
* A2 - Pointer to start of current video RAM line (Y2)
* A3 - Leftmost pixel X co-ordinate
* A4 - Background pixel colour, normalised
* A5 - Y step between lines: 128=CSIZE ?,0; 256=CSIZE ?,1
* A6 - Pointer to QDOS or Argos system variables
* A7 - SUPERVISOR stack pointer
*      
* First let's read one pixel from line 0 for reference
*
no_wider   move.l   d1,a3           Preserve character X edge
           moveq    #7,d2           Mask bottom 3 bits of X
           and.w    d1,d2           D2 := pixel # in word, 0-7
           lsr.w    #3,d1           D1 := line byte offset, X DIV 8
           add.w    d1,d1           Address only even words 0-126 
           move.w   0(a2,d1.w),d4   D4.W := background screen word
           lsl.w    d2,d4           Shift pixel data to high bits   
           and.w    d7,d4           Mask out a single pixel
           move.l   d4,a4           Save background for later
*
* Scan all nine lines and compare them with the STRIP colour in A4
*
           moveq    #8,d2           DBRA through 9 lines
scanline   add.l    a5,a2           Advance to next line
           move.w   a3,d1           Reset X to left edge
           moveq    #0,d6           Clear fount pattern
           moveq    #5,d3           DBRA count for X (5-0)
*
* Scan a row of six pixels. A2 -> video RAM line, D1 is pixel X, 0-511
*
scan_row   moveq    #7,d4           Mask for bottom bits of X
           and.w    d1,d4           X && 8 is pixel # in word
           move.w   d1,d0           Copy old X before massage
           lsr.w    #3,d0           Offset on line is X DIV 8
           add.w    d0,d0           Address only even words 0-126 
           move.w   0(a2,d0.w),d0   D0.W := relevant screen word
           lsl.w    d4,d0           Align pixel in Most Sig bits   
           and.w    d7,d0           Mask out the background
           cmp.w    a4,d0           Does pixel match the strip?
           beq.s    roll_on         If yes, leave bit unset  
           addq.w   #2,d6           Set one fount bit
roll_on    add.w    d6,d6           Shift pattern along
           add.w    d5,d1           Advance to next column
           dbra     d3,scan_row
           move.b   d6,(a1)+        Stuff fount line in buffer
           dbra     d2,scanline     
           lea.l    -9(a1),a1       Retrace to start of buffer
*
* Look up the character pattern at (A1) in both founts
*
lookup     moveq    #0,d1           Clear high bytes of code
           move.b   (a1)+,d4        Pre-read the first line
           move.l   42(a0),a2       A2 -> Fount 1 base
           bsr.s    scan_fount
           beq.s    result_d1       Match?
           move.l   46(a0),a2       A2 -> Fount 2 base
           bsr.s    scan_fount
           beq.s    result_d1
           moveq    #-7,d1          NOT FOUND error code
result_d1  moveq    #0,d0           Signal no QDOS error
           rts
*
* SCAN_FOUNT; corrupts D0, D1, D2, A2, A3, A4
* Entry: A2 -> Fount base, A1 -> Pattern+1, D4 = start of pattern
* Exit:  Z means no match, otherwise D1=ASCII code of pattern
*
scan_fount move.b   (a2)+,d1        Get first pattern # in fount
           moveq    #0,d0           DBRA pattern count in bytes
           move.b   (a2)+,d0        D0 := Number of patterns - 1
test_start cmp.b    (a2)+,d4        Exclude 86-88% of matches
           bne.s    zoom_on         Ignore the rest if mismatched
           moveq    #5,d2           DBRA count for more lines
           move.l   a1,a3           Scan pattern at (A3), saving A1
           move.l   a2,a4           Check versus (A4) preserving A2
scan_all   cmpm.b   (a4)+,(a3)+
           dbne     d2,scan_all     Check each line of pattern   
           beq.s    descenders      First 7 lines are OK...
zoom_on    addq.l   #8,a2           Skip remainder of pattern
           addq.b   #1,d1           Assume next ASCII code
           dbra     d0,test_start
           moveq    #-7,d0          Not found, exit with Z reset
           rts
*
descenders btst     #0,66(a0)       Check for UNDER 1
           bne.s    last_line
           move.b   (a4),d2         UNDER 0, Check line 8
           cmp.b    (a3),d2
           bne.s    zoom_on
last_line  move.b   1(a4),d2        Line 9 (last in pattern)
           cmp.b    1(a3),d2
           bne.s    zoom_on         So near and yet so far...
           rts                      MATCH!  Exit with Z set
*
* Routine names and addresses
*
define     dc.w     0,0             No procedures
           dc.w     2               Two functions
           dc.w     clip_int-*
           dc.b     5,'CLIP%'
           dc.w     clip_str-*
           dc.b     5,'CLIP$'
           dc.w     0               End of list
*
           end
