* QL WORLD DIY TOOLKIT - SET, SET# & ALTER extensions - RESIDENT VARIABLES
* Version 1.66, Copyright 1991 Simon N Goodwin, suggested by Luca Pivato.
*
lump_size  equ      1008              RAM lump size taken from Common Heap
call_code  equ      20153             680XX OPCODE for JSR .L
string     equ      1                 SuperBASIC type code of a string
float      equ      2                 Type code for floating-point values
integer    equ      3                 16 bit signed integer type code
*
start      lea.l    define,a1         Point to the table of details
           move.w   $110\w,a2         Find BP.INIT (a word vector)
           jmp      (a2)              Add SET to SuperBASIC
*
* Internal variables, not for ROM; could be made a QPAC or Argos THING
*
heap_lump  dc.w     lump_size         SET memory allocation size, 16-32760
sentinel   dc.l     0                 Heap pointer   (NOT suited to ROM!)
*
* These value-fetching routines are shared by all SET functions
*
read_int   moveq    #2,d1             Extra space needed
           bsr.s    checker           Allocate room on the stack
           move.l   (a7)+,a0          A0 -> Result
           move.w   (a0),0(a1,a6.l)   Fetch and stack the result
           moveq    #integer,d4       Indicate INT result
           rts                        Return D0 set by CHECKER
*
read_ptr   moveq    #6,d1             Extra space needed
           bsr.s    checker
           move.l   (a7)+,a0          A0 -> Result
           move.w   (a0),d3           Checked for even-ness earlier
           moveq    #0,d0             MT.INF trap key (be sure)
           trap     #1                Find the system variables
           move.l   0(a0,d3.w),d2     Read the system vector 
           move.l   d2,d1             D1 will be the exponent
           beq.s    normalised        Job done, if D2 & D1=0; result 0
*
* This alternative code normalises *typical* SYSVAR addresses extra-quickly!
* It replaces sequential shifts and checks with one fast shift if the value
* is in the first 64K of QL RAM, as is often true for system table pointers.
* It can be seven times faster than the binary method normally used in DIY
* Toolkit. The bitwise loop is about 30 per cent slower than the binary
* method, even though this version only works with positive numbers and
* grows especially slow on (most unlikely) small values, like 1.
*
*           swap     d1                Inspect the high word of D1
*           subq.w   #2,d1             Is D1 between 128K..192K-1 ?
*           bne.s    slow_loop         No, so normalise D2 slowly 
*           moveq    #13,d1            Bit Shift count for D2
*           lsl.l    d1,d2             Normalise D2 in one step!
*           move.w   #2066,d1          Set exponent for 131072-262143
*           bra.s    normalised
*
slow_loop  move.w   #2080,d1          Guess at the exponent + 1
normaloop  subq.w   #1,d1             Halve the weight of the guess
           add.l    d2,d2             Double the mantissa
           bpl.s    normaloop         Does it still fit?
           lsr.l    #1,d2             Whoops, ensure sign=0 (+ve)    
normalised move.w   d1,0(a1,a6.l)     Exponent
           move.l   d2,2(a1,a6.l)     Mantissa
           moveq    #float,d4         Indicate FLOAT result
           rts
*
read_float moveq    #6,d1             Extra space needed
           bsr.s    checker
           move.l   (a7)+,a0          A0 -> Result
           move.w   (a0)+,0(a1,a6.l)
           move.l   (a0),2(a1,a6.l)   Transfer mantissa
           moveq    #float,d4         Floating-point result
           rts
*
read_str   move.l   (a7)+,a0          A0 -> Address of Result (length.W)
           move.l   (a0),a0           Pick up string value address
           moveq    #3,d1             Room for length & odd byte
           add.w    (a0),d1           Add room needed for text
           bclr     #0,d1             Count in whole words
           bsr.s    checker           Check there's room
           lsr.w    #1,d4             D4 := D4 DIV 2, # text words
           subq.w   #1,d4             Adjust word count for DBRA
           move.l   a1,a2             A2 is the offset of stacked words
copy_text  move.w   (a0)+,0(a2,a6.l)  Stack one word from the heap
           addq.l   #2,a2             Advance up the maths stack
           dbra     d4,copy_text      Copy words till all are done
got_string moveq    #string,d4        Result datatype is STRING
           rts
* 
* Check for D1.L RI Stack bytes; alters D1-3, D4=old D1, A1=RI.SP
*
checker    move.l   d1,d4             Save size for use later
           tst.b    $54(a6)           Turbo/Supercharged code?
           bmi.s    found_room        Always 120+ bytes free
           move.w   $11A\w,a1         BV.CHRIX checks space
           jsr      (a1)              D1 bytes are needed
found_room move.l   $58(a6),a1        A1 := (new?) BV.RIP
           sub.l    d4,a1             Grab the bytes
           move.l   a1,$58(a6)        Update BV.RIP to suit
           moveq    #0,d0
           rts
*
* {SET [#] unset_name / ALTER "name_text"} TO value - parse parameters
*
alter      moveq    #-1,d5            If D5.L is -ve, it's ALTER
           bra.s    parameters
set        moveq    #1,d5             Distinguishing mark
parameters lea.l    16(a3),a4         Check for 2 parameters
           cmp.l    a4,a5
           bne.s    bad_param2
*
* Make A4 -> SuperBASIC (task 0,0) for Master Name Table access, etc.
*
           moveq    #0,d2             Search entire task tree
           moveq    #0,d1             Look for SuperBASIC
           moveq    #2,d0             MT.JINF Trap key
           trap     #1                A0 := base of task 0,0
           move.l   a0,a4
           tst.l    d5                Is this ALTER?
           bmi.s    lookup
*
* D5.B is the type to be SET, or 0 for VECTOR type. This code does
* not NEED to support three types - it could have strings all through -
* but it is much faster and probably more concise with INTS & FLOATS.
*
           move.b   1(a3,a6.l),d5     Get first NT entry type
           beq.s    bad_param2        Strange, and best avoided
           btst     #7,d5             Is there a # at the start?
           beq.s    normal_set
           moveq    #0,d5             Flag a special SET data type
normal_set move.b   0(a3,a6.l),d1     Get the name type too
           beq      notyetset         Handle an unset name
*
bad_param2 moveq    #-15,d0
bad_exit2  rts
*
* It is not a name, so try to evaluate parameter 1 as a string
*
lookup     lea.l    8(a3),a5          Isolate the first parameter
           move.l   a5,d7             Remember what we forget for now 
           move.w   $116\w,a0         Fetch CA.GTSTR vector
           jsr      (a0)              Try to fetch a string parameter
           bne.s    bad_exit2         Exit unless it worked
           move.w   0(a1,a6.l),d0
           beq.s    bad_param2        Reject a null string
           move.w   d0,d1             Save length for later
           lea.l    2(a1),a5          Save offset of text
*
* Convert lower to UPPER case; reset bit 5 of parameter bytes
*
           moveq    #-33,d2           Conversion mask = NOT $20
lock_case  and.b    d2,2(a1,a6.l)
           addq.l   #1,a1             Advance through text
           subq.w   #1,d0             Count down length
           bne.s    lock_case         Convert all characters
*
* Work out the datatype to be SET from the name text, if need be
*
           tst.b    d5                Do we need the type of parameter 1?
           beq.s    type_cast         No, it's a float-type SET # command
           move.b   1(a1,a6.l),d3     Get the last character of the name
           moveq    #string,d5        Assume a string, to begin with
           cmp.b    #'$'-32,d3        Did the name end with a dollar?
           beq.s    type_cast         Yes, it's a string
           moveq    #integer,d5       Maybe an integer, then?
           cmp.b    #'%'-32,d3        Did it end with a per cent sign?
           beq.s    type_cast         It's an integer
           moveq    #float,d5         Other last characters denote floats        
*
* Now try to find the name text in the SuperBASIC Name List
* 
type_cast  trap     #0                Pre-empt disturbances
           move.l   24(a4),a0         A0 -> Name Table Start
           move.l   28(a4),d0         D0 -> Name Table End
           move.l   32(a4),d3         D3 -> Name List Start
*
next_name  move.w   2(a0,a4.l),a3     A3.L is offset in NL
           adda.l   d3,a3             (A3,A4.L) -> Name
           cmp.b    0(a3,a4.l),d1     Compare length
           beq.s    got_length        Length matches!
advance_nl addq.l   #8,a0             Advance through NL
           cmp.l    a0,d0             Stop at the end
           bhi.s    next_name
*
* The parameter name is not in the Name List! THEORY: Create
* the required heap entry, then use BP.INIT to instantiate
* the name. This is not enough to help a compiled tasks, which
* only look up names once, upon loading. The BASIC pre-loader
* must use SET to create the name entry; later ANY TASK can ALTER
* its value if they want; all changes are globally accessible.
*
* This routine to call BP.INIT to create names is not implemented
* because it seems overcomplicated and would allow SET <any string
* expression> TO <any apt value> - this seems good till you realise
* that accidental repeated SETs will CREATE RESIDENT FUNCTIONS using
* NAMES that correspond to the VALUES of the re-SET variables.
* I decided I could live without that!
*
* If you're keen, here's a general plan for adding that 'feature':
*
*          Stack the pointer to the string
*          Supply a 'dummy' pointer into BUFFER; MOVE.L (A6),D7 (aha!) 
*          Call as if the NT entry was found
*          Pick up code address vector from 4(A6,D7.L) @ BUFFER
*          Copy from the string to a BP.INIT buffer on A7...
*
*          DC.L 0                         No procedures
*          DC.W 1                         1 function, please
*          DC.W (length+1) div 8+1
*          DC.B length,'name'
*          DC.W 0                         Word boundary; End of FNs
*
*          Logical checks - length is 1-60 (say; needs 70 SP bytes, ok!?)
*                         - exclude code 0s (spaces & nulls) to trap crap
*                         - QVIEW reject most other codes; QDOS permits them
*
* Call BP.INIT last, to avoid A6/A4/D4 disruption (MG quirk, & Minerva?)
*
           moveq    #-7,d0            Not found (yet)
           bra      super_stop
*
*
* Check the name to see if it matches the parameter
*
got_length move.b   1(a3,a4.l),d4
           and.b    d2,d4             Ensure consistent case
           cmp.b    0(a5,a6.l),d4
           bne.s    advance_nl        Mismatch, try another
           move.w   d1,d6             Save residual length
           subq.w   #2,d6             DBRA count the rest
           bmi.s    found_it          Found name, length 1
           move.l   a5,a1             D4 & A1 are temporaries
check_name move.b   2(a3,a4.l),d4
           and.b    d2,d4             Convert case of name
           addq.l   #1,a3             Step through Name List
           addq.l   #1,a1             Step through parameter
           cmp.b    0(a1,a6.l),d4
           dbne     d6,check_name
           bne.s    advance_nl        Name mismatch, go on
found_it   cmp.b    #9,0(a0,a4.l)
           beq.s    lookup_ok         It's already a resident FN, hurrah
in_use     moveq    #-9,d0            Report that the name is 'in use'
           bra      super_stop
*
lookup_ok  andi     #$D8FF,sr         User mode re-starts the scheduler
           move.l   d7,a3             D7 points past parameter 1
           lea.l    8(a3),a5          Now A3 & A5 bracket parameter 2
           move.l   a0,d7             D7 is the offset of the NT entry
           bra.s    got_name    
*
bad_param  moveq    #-15,d0           ERR.BP report code
bad_exit   rts                        Return error code in D0
*
notyetset  move.w   2(a3,a6.l),d7     Get parameter name NT index
           ble.s    bad_param
           ext.l    d7                
           lsl.l    #3,d7             Scale for 8 byte NT entries
           add.l    24(a6),d7         Add offset from Basic base to NT
next_param addq.l   #8,a3             Advance to the next parameter     
*
* A4 -> SuperBasic, A3 & A5 -> Parameter #2, D7 is Basic NT entry offset
* D5.B is the data-type, 1-3, or 0 for a 'vector'
* Evaluate the expression at (A3,A5), type D5.B, onto the Maths Stack
*
got_name   and.b    #15,d5            Isolate documented data-type bits
           move.b   d5,d4             Make a temporary copy of the type
           bne.s    not_vector        Type 0 in D4/5 means it's a vector
           moveq    #3,d4             Integer parameters suit SET #
not_vector ext.w    d4                Type Word := 1 , 2 , 3
           add.w    d4,d4             Now type code is 2, 4 or 6.W
           suba.l   a0,a0             Clear A0 the quick way
           suba.w   d4,a0             Remember the implicit EXT.L
           movea.w  $118(a0),a0       Pick up a vector: $116\114\112?
           jsr      (a0)              Put parameter value on stack
           bne.s    bad_exit
*
* Is it a pre-existing function?
*
           tst.b    0(a4,d7.l)        Has the name got a value yet?
           beq.s    neophyte          Name types 0 & 9 are possible
           move.l   4(a4,d7.l),a0     A0 is the code vector
           cmp.w    #call_code,(a0)+  Is it one of mine (probably!)
           bne.s    bad_param         Only change FNs that start right 
           cmp.b    #1,d5             Re-defining a preset string?
           bne.s    old_scalar
*
* It's a string; deallocate and re-allocate space IF NECESSARY...
*
           lea.l    4(a0),a3          Make A3 point at the data handle
           move.l   (a3),a0           Find the current data area
           move.l   -4(a0),d1         Find maximum space for text here
           subq.l   #6,d1             Ignore both .L & .W prefixes
           cmp.w    0(a1,a6.l),d1     Will the altered text fit inside?
           bcc.s    str_fits
*
* De-allocate the old space and find some anew
*         
           trap     #0                Prevent scheduling of other tasks
           move.w   #16,-8(a3)        Shrink the descriptor
           addq.l   #6,d1             Don't forget room for the lengths
           move.l   a3,d7             Preserve address of the data pointer
           subq.l   #4,a0             Step back to heap block length
           lea.l    sentinel,a1       Release to my heap sentinel
           move.w   $DA\w,a2          Get MM.LNKFR vector
           jsr      (a2)              Release it!
           moveq    #1,d1             An odd byte, to get the size started
           moveq    #-1,d5            Special type for ALTERed strings
           move.l   $58(a6),a1        Find the parameter stack again
           bra.s    new_string        Store the ALTERed text
*
str_fits   moveq    #1,d4             Allow for an odd byte length
           add.w    0(a1,a6.l),d4
           lsr.w    #1,d4             Count words for DBRA
           bra.s    re_set
*
* Plug the new value into the scalar slot... 
*
old_scalar addq.l   #4,a0             Skip the code address (check it?)
           moveq    #2,d4             Presumed length 3 words for DBRA
           cmp.b    #float,d5         Is it a float?
           beq.s    re_set
           moveq    #0,d4             No, just an integer
re_set     trap     #0                Stop over-eager tasks misreading
           bra      data_store
*
neophyte   trap     #0                Prevent multi-tasking temporarily
*
*   find room for the new value
*   generate code
*   connect the NT entry at (A4,D7.L) to the code
*   return
*
no_release moveq    #16,d1            Size = 4(LEN) + 6(JMP) + 6?(DATA)
           cmp.b    #string,d5        Strings have special requirements
           beq.s    new_string
           cmp.b    #float,d5         Floating point? (6 bytes DATA)
           beq.s    space_set         Yes, D1 was guessed right
           moveq    #12,d1            No, integer, D1 should be 4+6+2
           bra.s    space_set
*
new_string addq.l   #6,d1             Allow for string spacer & long header
           add.w    0(a1,a6.l),d1     Still works if total overflows 32K
space_set  move.l   d1,d4             Save the required size for later
alloc_d1   lea.l    sentinel,a0       Find the start of the user heap
           move.w   $D8\w,a2          Get MM.ALLOC vector
           jsr      (a2)              Find D1 bytes
           tst.l    d0                N.B. Explicit test NECESSARY here
           beq.s    count_up          Set them, if they were found
*
* Find more room for data in my heap; clobbers D0-3, A0-3
*
find_more  move.w   heap_lump,a0      Implicitly EXT.L A0
           move.l   d4,d1             D1 is space required
           cmp.l    a0,d1             Will it fit the next lump?
           bhi.s    get_plenty        If not, get room PLUS a lump
           moveq    #0,d1             No extra needed...
get_plenty add.l    a0,d1             Find the space needed
           moveq    #0,d2             Get RAM owned by SuperBASIC
           moveq    #24,d0            Set MT.ALCHP trap key
           trap     #1                Ask for Common Heap space
           tst.l    d0                Did we get it?
           bne      super_stop
           moveq    #16,d2            Forget the 16 byte ALCHP header
           sub.l    d2,d1             D1 is the size of the available space
           lea.l    sentinel,a1
           move.w   $DA\w,a2          Get MM.LNKFR vector
           jsr      (a2)              Link the new space into the heap
           move.l   d4,d1             Remember what we originally wanted
           bra.s    alloc_d1          Go back for it
*
* Count data words to be copied from RI stack to my heap, and make code
*
bad_vector moveq    #-15,d0           Reject odd vector parameter values 
           bra.s    super_stop
*
make_ptr   btst     #0,1(a1,a6.l)     Check it's an even word!
           bne.s    bad_vector
           lea.l    read_ptr,a2
           bra.s    word_count
*
make_nums  lea.l    read_int,a2       Point to integer code
           cmp.b    #integer,d5       But is it an integer?
           beq.s    word_count        Phew...
           lea.l    read_float,a2     No, FLOAT is Hobson's choice
           bra.s    word_count
*
count_up   move.l   $58(a6),a1        Find the parameter stack again
           tst.b    d5                Identify revived strings now
           bpl.s    make_code         Create the code for a new name
           move.l   d7,a2             A2 -> code handle to string text
           addq.l   #4,a0             Step past the extension entry length
           move.l   a0,(a2)           Make the handle point at the new text
           lsr.w    #1,d4             Prepare to count in words
           subq.l   #3,d4             Discard 2 prefix words & 1 for DBRA
           bra.s    data_store
*
make_code  cmp.b    #string,d5        Check the data-type
           bcs.s    make_ptr          If less than one presume 0, VECTOR
           bhi.s    make_nums         It's higher than 1, FLOAT or INTEGER
           lea.l    read_str,a2       Point to string-fetching code
           sub.w    #9,d4             Round up & forget extra string data
word_count lsr.w    #1,d4             Count words not bytes
           subq.w   #6,d4             Discount 5 prefix words + 1 for DBRA
*
* Generate code and data 
*
make_value move.l   (a0)+,d1          Recall the total heap block length
           move.l   a0,4(a4,d7.l)     Set a Task 0 Name Table code pointer
           move.w   #call_code,(a0)+  Store a JSR.L instruction
           move.l   a2,(a0)+          Store the address of the data fetcher
           move.w   #$0900,0(a4,d7.l) Mark this Name as a Resident Function
           subq.b   #string,d5        Is this a string?
           bne.s    data_store        If not, we have almost finished
           lea.l    10(a0),a2         Find address of string (4+2+4 later)
           move.l   a2,(a0)+          Put it after the call
           addq.l   #2,a0             Skip to the next user heap block
           moveq    #16,d2            Size of the first part
           sub.l    d2,d1             D1 is size of the remaining part
           move.l   d1,(a0)+          Store the block size of part 2
data_store move.w   0(a1,a6.l),(a0)+  Copy data words to my heap
           addq.l   #2,a1             Advance up this task's RI stack 
           dbra     d4,data_store     Count the data word(s)
super_stop andi     #$D8FF,sr         Return to multi-tasking user mode 
           rts
*
define     dc.w     2                 Two procedures
           dc.w     set-*
           dc.b     3,'SET'
           dc.w     alter-*
           dc.b     5,'ALTER'
           dc.w     0,0,0             End of Procs; no Functions (yet!)
*
           end
