* Sinclair QL World DIY TOOLKIT - WINDOWS for QDOS
* Version 1.2m, Copyright 1992 Simon N Goodwin
* Updated September 1993 for Metacomco Assember 1.7
*
start       lea.l      define,a1      Point at names & addresses
            movea.w    $110.w,a2      Read the BP.INIT vector
            jmp        (a2)           Link SuperBasic extensions
*
* base_address = W_CRUNCH ( #ch% , flag% [0=All 2=Red 4=Green] )
*
crunch      bsr        get_channel    Pick up channel number
            subq.w     #2,d3          Check parameter count
            bne        bad_param2     Two are expected
            move.w     2(a1,a6.l),d6  Pick up compression flag
            addq.l     #2,$58(a6)     Tidy RI stack pointer
            bra.s      save_window
*
* base_address = W_STORE ( #ch% )     Store window contents
*
store       bsr        get_channel    Pick up channel number
            subq.w     #1,d3          Check parameter count
            bne        bad_param2     Only one is expected
            moveq      #0,d6          Flag for no compression
*
save_window bsr        find_window    Put details in registers
            move.w     d0,d7          Save width for later use
            moveq      #-1,d2         RAM owner is this task
            moveq      #24,d0         MT.ALCHP trap key
            trap       #1             Allocate Common Heap RAM
            tst.l      d0             Did that work?
            bne.s      exit           Return error code
*
            move.l     a0,d1          Preserve address of RAM
            move.w     d7,d0          Retrieve line word count          
            move.l     #'Bufa',-4(a0) Store flag for DISCARD
            move.l     224(a6),(a0)+  Point on to the next ALCHP
            move.l     d1,224(a6)     Extend the ALCHP list
*
            move.l     d5,(a0)+       Record total space used
            move.w     d4,(a0)+       Record line count
            move.w     d6,(a0)+       Record compression flag
* 
* A4 -> First video word, A0 -> Buffer for screen information
* D4.W is count of pixel lines-1, D0.W is line width in words-1
*
            subq.w     #2,d6          Check the CRUNCH flag
            bmi.s      save_lines     If <2, store both bytes
            bne.s      crunch_line    >2, store green bytes only
            addq.l     #1,a4          Point at first red byte
crunch_line lea.l      save_bytes,a3  Point at byte saver
            bra.s      save_it
save_lines  lea.l      save_words,a3  Point at word saver
save_it     bsr        scan_lines     Call (A3) for each line
*
* RETURN_FP stacks D1.L in SuperBasic floating-point form
*
return_fp   move.w   d1,d4            D4 will be exponent
            move.l   d1,d5            D5 will be mantissa
*           beq.s    normal           A most unlikely case!
            move.w   #2079,d4         First guess at exponent
            add.l    d1,d1            Already normalised?
            bvs.s    normal
            subq.w   #1,d4            No, halve exponent weight
            move.l   d1,d5            Double mantissa to match
            moveq    #16,d0           Try a 16 bit shift
normalise   move.l   d5,d1            Take copy of mantissa
            asl.l    d0,d1            Shift mantissa D0 places
            bvs.s    too_far          Overflow; must shift less
            sub.w    d0,d4            Correct exponent for shift
            move.l   d1,d5            New mantissa is more normal
too_far     asr.w    #1,d0            Halve shift distance
            bne.s    normalise        Try shift of 8, 4, 2 and 1
*
* Check there's six bytes of space for the result
*
normal      moveq    #6,d1            Number of bytes needed
            movea.w  $11A.w,a0        BV.CHRIX vector
            jsr      (a0)
            movea.l  $58(a6),a1       Get a safe A1 value
            subq.l   #6,a1
            move.l   a1,$58(a6)       Grab 6 more bytes
            move.l   d5,2(a1,a6.l)    Stack mantissa
            move.w   d4,0(a1,a6.l)    Stack exponent
            moveq    #2,d4            Floating point result
            moveq    #0,d0
exit        rts
*
* GET_CHANNEL - evaluates integer channel number parameter
* Returns ID in A0 or returns error code to 4(A7)
*
get_channel movea.w    $112.w,a2      CA.GTINT - get integers
            jsr        (a2)
            bne.s      quick_exit     Return D0 error code
            move.w     0(a1,a6.l),d7  Pick up the channel #
            bmi.s      bad_param      It should be positive
            addq.l     #2,$58(a6)     Update BV.RIP
default     mulu       #40,d7         Find offset in table
            add.l      48(a6),d7      Add base to offset
            cmp.l      52(a6),d7      Check not beyond end
            bge.s      chan_error     Number is too high
            move.l     0(a6,d7.l),d7  ID for output window
            bmi.s      chan_error     Channel is closed
chan_open   movea.l    d7,a0          Set up Channel ID
            rts
*
* W_SWAP #ch% - exchanges red & green display bytes
*
swop        bsr.s      get_channel    Read channel number
            subq.w     #1,d3          Check parameter count
            bne        bad_param2     Only one is expected
            bsr.s      find_window
            lea.l      swop_words,a3
*
scan_lines  lea.l      128(a4),a2     Point at the next line
            move.w     d0,d2          Duplicate width for loop
            jsr        (a3)           Call chosen line scanner
            movea.l    a2,a4
            dbra       d4,scan_lines  Process all the lines
            moveq      #0,d0          It worked!
            rts
*
* SET_RED #ch% , flag%  [1=Set 0=Clear -1=Copy to red bytes]
*
setred      bsr.s      get_channel    Read channel number
            subq.w     #2,d3          Check parameter count
            bne.s      bad_param2     Two are expected
            move.w     2(a1,a6.l),d6  Fetch the flag
            bsr.s      find_window
            tst.w      d6
            bmi.s      red_lines      Copy green bytes to red
            addq.l     #1,a4          Point at odd (red) bytes
*
check_pat   beq.s      fill_lines     If D6=0, leave it alone
            moveq      #-1,d6         Otherwise, set all bits
fill_lines  lea.l      fill_bytes,a3  Put D6 in alternate bytes
            bra.s      scan_lines
red_lines   lea.l      red_bytes,a3   Point at next line
            bra.s      scan_lines
*
bad_param   moveq      #-15,d0        BAD PARAMETER error
            bra.s      quick_exit
chan_error  moveq      #-6,d0         CHANNEL NOT OPEN error
quick_exit  addq.l     #4,a7          Tidy the return stack
            rts                       Return subroutine errors
*
* SET_GREEN #ch% , flag%  [1=Set 0=Clear -1=Copy to green bytes]
*
setgreen    bsr.s      get_channel    Read channel number
            subq.w     #2,d3          Check parameter count
            bne.s      bad_param2     Only one is expected
            move.w     2(a1,a6.l),d6
            bsr.s      find_window
            tst.w      d6
            bpl.s      check_pat      Set or clear even bytes
            addq.l     #1,a4          Start with the red byte
            lea.l      green_bytes,a3 Point at byte copier
            bra.s      scan_lines
*
* FIND_WINDOW - locates window details from channel definition
*
find_window lea.l      window_ext,a2  Point A2 at EXTOP code
            moveq      #-1,d3         Wait indefinitely long
            moveq      #9,d0          SD.EXTOP trap key
            trap       #3             Call via Qdos
            tst.l      d0             Z signals ERR.OK
            bne.s      quick_exit     Return error directly
            move.w     d1,d0          D0 is window line count
            move.l     d1,d4          Preserve width & height
            movea.l    a1,a4          Preserve base address
            swap       d1             Recover line width
            mulu       d0,d1          Compute storage needs
            tst.w      d6             Are we crunching?
            beq.s      no_crunch
            lsr.l      #1,d1          Halve space needed
no_crunch   addq.l     #6,d1          Allow for size & depth
            addq.l     #6,d1          Allow for link & flag
            move.l     d1,d5          Preserve size needed
            swap       d4
            move.w     d4,d0          Retrieve line width
            lsr.w      #1,d0          Count words not bytes
            subq.w     #1,d0          Adjust for DBRA later
            swap       d4
            subq.w     #1,d4          Adjust for DBRA
            rts
*
bad_range   moveq      #-4,d0         OUT OF RANGE error code
            rts
bad_param2  moveq      #-15,d0        BAD PARAMETER error code
exit2       rts
*
* W_SHOW #ch% , base_address          Re-displays stored windows
*
show        lea.l      16(a3),a4      Point A4 past parameter 2
            cmpa.l     a4,a5          Are there 2 parameters?
            bne.s      bad_param2
            subq.l     #8,a5          Forget the first for now
            bsr        get_channel
            movea.l    a5,a3          Retrieve parameter 2
            addq.l     #8,a5
            movea.w    $118.w,a2      CA.GTLIN, get a long word
            jsr        (a2)
            bne.s      exit2          Return D0 error code
            move.l     0(a1,a6.l),d7  Pick up the base address
            bmi.s      bad_param2     It should be positive
            btst       #0,d7          and even
            bne.s      bad_param2
*
* Now D7 is the buffer base address and A0 is the channel ID
*
            movea.l    d7,a2          A2 -> RAM buffer
            move.w     10(a2),d6      Check compression flag         
            bsr.s      find_window    
            movea.l    d7,a0          A0 -> RAM buffer
            addq.l     #4,a0          Skip the ALCHP link
            cmp.l      (a0)+,d5       Check total space used
            bne.s      bad_range
            cmp.w      (a0)+,d4       Check line count
            bne.s      bad_range
            addq.l     #2,a0          Skip the CRUNCH flag now
* 
            subq.w     #2,d6          Check the CRUNCH flag
            bmi.s      word_lines     If <2, restore both bytes
            bne.s      byte_lines     >2, set green bytes only
            addq.l     #1,a4          Point at first red byte
byte_lines  lea.l      load_bytes,a3
            bra        scan_lines
word_lines  lea.l      load_words,a3
            bra        scan_lines
*
* Find SCR_BASE (A1.L) LINE_WIDTH% (D1.H) LINE_COUNT% (D1.W)
*
window_ext  moveq      #40,d0         Preset MEM length
            cmp.l      (a0),d0        Check this channel
            beq        bad_param2     Reject MEM definitions
            move.w     24(a0),d0      Fetch margin in pixels
            lsr.w      #2,d0          Divide by 4
            bclr       #0,d0          Make even word offset
*
            movea.l    50(a0),a1      Locate the screen start
            adda.w     d0,a1          Adjust for left margin
            move.w     26(a0),d1      Fetch top margin
            mulu       #128,d1        Scale for 128 bytes/line
            adda.l     d1,a1          A1 -> first window word
*
            moveq      #6,d1          Round to word boundaries
            add.w      24(a0),d1      Allow for left margin
            add.w      28(a0),d1      Allow for width 
            lsr.w      #2,d1          Divide width by 4
            bclr       #0,d1          Make word offset even
            sub.w      d0,d1          D1 is line width in bytes
*
            swap       d1             Put byte width in D1.H
            move.w     30(a0),d1      Put line count in D1.L
            moveq      #0,d0          Indicate no error
            rts
*
* Line scanning routines, used by JSR (A3) in SCAN_LINES
*
save_bytes  move.b     (a4)+,(a0)+    Copy from screen to RAM
            addq.l     #1,a4          Skip the other byte
            dbra       d2,save_bytes  Process one window line
            rts
*
fill_bytes  move.b     d6,(a4)        Copy pattern to byte
            addq.l     #2,a4          Advance to next word
            dbra       d2,fill_bytes  Scan one window line
            rts
*
green_bytes move.b     (a4),-(a4)     Copy red byte to green
            addq.l     #3,a4          Point at next red byte
            dbra       d2,green_bytes Scan one window line
            rts
*
red_bytes   move.b     (a4)+,(a4)+    Copy 1 green byte to red
            dbra       d2,red_bytes   Scan the window line
            rts
*
load_bytes  move.b     (a0)+,(a4)+    Copy from heap to screen
            addq.l     #1,a4          Skip the other byte
            dbra       d2,load_bytes  Process one window line
            rts
*
load_words  move.w     (a0)+,(a4)+    Copy from RAM to screen
            dbra       d2,load_words  Process one window line
            rts
*
swop_words  move.b     (a4)+,d1       Fetch green byte
            move.b     (a4),-1(a4)    Copy red to green
            move.b     d1,(a4)+       Store green over red
            dbra       d2,swop_words  Scan one window line
            rts
*
save_words  move.w     (a4)+,(a0)+    Copy from screen to RAM
            dbra       d2,save_words  Process one window line
            rts
*
define      dc.w       5              5 new PROCedures
            dc.w       show-*
            dc.b       6,'W_SHOW'
            dc.w       swop-*
            dc.b       6,'W_SWOP'
            dc.w       swop-*
            dc.b       6,'W_SWAP'
            dc.w       setred-*
            dc.b       7,'SET_RED'
            dc.w       setgreen-*
            dc.b       9,'SET_GREEN'
            dc.w       0              End of PROCs
*
            dc.w       2              2 new FuNctions
            dc.w       store-*
            dc.b       7,'W_STORE'
            dc.w       crunch-*
            dc.b       8,'W_CRUNCH'
            dc.w       0              End of FNs
*
            end
